var size_x = 30;
var size_y = 25;

var lives = 4;
var toEat = 0;

var dirs = {"w": [0, -1], "n": [-1, 0], "e": [0, 1], "s": [1, 0]};

var smileys = {default: "😀", fed: "😋", panic: "😱", dead1: "💀", dead2: "⚰️", cool: "😎", rofl1: "🙃", rofl2: "🙂", rofl3: "🤣", party: "🥳"};

var symbols = ["😀", "😋", "😱", "💀", "⚰️", "😎", "🙃", "🙂", "🤣", "🥳", "👻", "👀", "🔥", "💧", "•", "🔸", "🌶️", "🍬", "🍭", "🍫",
	       "🍎", "🍓", "🍒", "🍪", "🌀", "🍧", "🍏", "🧯"];

var wallchars = "wd"

var difficulty = "normal";
var baseSpeed = 30;
var gameTimer = null;
var ticks = 0;
var current_level = 0;

var rofl_count, rofl_timer;

var smileMan;

var music = false;
var sound = false;

var events = {};

var score = 0;
var highscore = 0;
var score_factor = 1.0;
var reachable = [];
var pathways = [];

var tunes = {
    happy: "RetRMN_Town.ogg",
    sad: "RetRMN_Dungeon.ogg",
    demo1: "RetRMN_MissingFound.ogg",
    demo2: "RetRMN_WishingFairy.ogg"
};
var sounds = {
    eat: "pacman_chomp.ogg",
    fruit: "pacman_eatfruit.ogg",
    eat_monster: "pacman_eatghost.ogg",
    bonus: "tm2_chime002.ogg",
    power: "tm2_metalhit001.ogg",
    monster: "tm2_laser001.ogg",
    caught: "tm2_bom002.ogg",
    death: "pacman_death.ogg",
    extra: "pacman_extrapac.ogg",
    level: "pacman_intermission.ogg",
    victory: "pacman_beginning.ogg"
};
var music_objects = [];
var sound_objects = {};
var current_tune = null;
var active_monsters = [];
var the_flames = null;
var music_volume = 0;
var sound_volume = 0;

var monster_types = {
    "blue": {pic: "ghost-blue", ai: "patrol"},
    "green": {pic: "ghost-green", ai: "hunt"},
    "red": {pic: "ghost-red", ai: "chase"},
    "yellow": {pic: "ghost-yellow", ai: "hunt"},
    "orange": {pic: "ghost-orange", ai: "hunt2"},
    "pink": {pic: "ghost-pink", ai: "chase2"},
};

var mappiece_types = {
    "d": {type: "wall"},
    "w": {type: "wall"},
    ".": {pic: "•",  type: "food", value: 10},
    "o": {pic: "🔸", type: "power", value: 100, power: 100}
};
var bonus_types = {
    "candy":     {pic: "🍬", type: "food", value: 50},
    "lolly":     {pic: "🍭", type: "food", value: 60},
    "cookie":    {pic: "🍪", type: "food", value: 80},
    "choclate":  {pic: "🍫", type: "food", value: 100},
    "grapes":    {pic: "🍓", type: "food", value: 200},
    "apple":     {pic: "🍎", type: "food", value: 300},
    "cherry":    {pic: "🍒", type: "food", value: 500},
    "icecream":  {pic: "🍧", type: "food", value: 800},
    "chili":     {pic: "🌶️", type: "power", value: 1000, power: 200},
    "turbo":     {pic: "🌀", type: "power", value: 2000, power: 100, speedup: 50},
};

function choice (array) {
    return array[Math.floor(Math.random() * array.length)];
}

function pct (p) {
    return Math.random () < (p/100);
}

function d (casts, sides) {
    let ret = 0;
    for (let i = 0; i < casts; i++) {
	ret += 1 + Math.floor (Math.random () * sides);
    }
    return ret;
}

function dx_mod (o1, o2) {
    return (size_x+o2.xPos-o1.xPos) % size_x;
}

function dy_mod (o1, o2) {
    return (size_y+o2.yPos-o1.yPos) % size_y;
}

function dx (o1, o2) {
    return o2.xPos-o1.xPos;
}

function dy (o1, o2) {
    return o2.yPos-o1.yPos;
}

function distance (o1, o2) {
    return Math.abs (dx(o1,o2)) + Math.abs (dy(o1,o2));
}

function opposite_of (dir) {
    return {w:"e", n:"s", e:"w", s:"n"}[dir];
}

function can_see (o1, o2) {
    return false;
}

function restart_effect (snd) {
    sound_objects [snd].currentTime = 0;
    effect (snd);
}

function effect (snd, callback=null, delay=0) {
    if (sound) {
	if (callback) {
	    sound_objects [snd].onended = callback;
	}
	sound_objects [snd].volume = sound_volume/100;
	sound_objects [snd].play ();
    } else if (callback) {
	setTimeout (callback, delay);
    }
}

var ais = {
    "patrol": patrol,
    "hunt": hunt,
    "hunt2": hunt2,
    "chase": chase,
    "chase2": chase2
};

function patrol (monster) {
    if (monster.dir && pct(50)) {
	let walls = tile (monster.xPos, monster.yPos).data ("walls");
	if (!walls || !walls.includes (monster.dir)) {
	    return;
	}
    }

    let choices = monster.choices ();
    if (!choices) {
	return;
    }
    choices = choices.filter (dir => dir != opposite_of (monster.dir));
    monster.dir = choice (choices);
}

function hunt (monster, fallback=patrol) {
    let dir;
    if (!dx (smileMan, monster)) {
	if (dy_mod (smileMan, monster) < size_y/2) {
	    dir = "n";
	} else {
	    dir = "s";
	}
    } else if (!dy (smileMan, monster)) {
	if (dx_mod (smileMan, monster) < size_x/2) {
	    dir = "w";
	} else {
	    dir = "e";
	}
    } else {
	fallback (monster);
	return;
    }
    let x = monster.xPos;
    let y = monster.yPos;
    while (true) {
	x = move_x (x, dir);
	y = move_y (y, dir);
	if (x == smileMan.xPos && y == smileMan.yPos) {
	    monster.dir = dir;
	    return;
	} else {
	    let p = tile (x, y).data ("piece");
	    if (p && p.type == "wall") {
		fallback (monster);
		return;
	    }
	}
    }
}

function hunt2 (monster) {
    if (distance (monster, smileMan) < 20) {
	chase2 (monster);
    } else {
	patrol (monster);
    }
}

function chase (monster) {
    try {
	let pw = pathways [monster.yPos][monster.xPos][smileMan.yPos][smileMan.xPos];
	if (pw [1] < 25) {
	    monster.dir = pw [0];
	} else {
	    hunt (monster);
	}
    } catch (e) {
	console.log ("chase::error", e, monster, smileMan);
	hunt (monster);
    }
}

function chase2 (monster) {
    try {
	let choices = [];
	let dist = pathways [monster.yPos][monster.xPos][smileMan.yPos][smileMan.xPos][1];
	for (dir of tile (monster.xPos, monster.yPos).data ("passages")) {
	    let x = move_x (monster.xPos, dir);
	    let y = move_y (monster.yPos, dir);
	    let d2 = pathways [y][x][smileMan.yPos][smileMan.xPos][1];
	    if (d2 < dist && !tile (x, y).data ("actor")) {
		choices.push (dir);
	    }
	}

	if (choices.length) {
	    monster.dir = choice (choices);
	} else {
	    hunt (monster);
	}
    } catch (e) {
	console.log ("chase2::error", e, monster, smileMan);
	hunt (monster);
    }
}

function add_score (delta) {
    score += Math.floor (score_factor*delta);
    $("#uscore").text (("00000000" + score).slice (-8));
    if (score > highscore) {
	highscore = score;
	window.localStorage.setItem ("smiling_highscore", highscore);
	$("#hiscore").text (("00000000" + highscore).slice (-8));
    }
}

function init_score () {
    score = 0;
    $("#score").text (("00000000" + score).slice (-8));
    highscore = window.localStorage.getItem ("smiling_highscore");
    $("#hiscore").text (("00000000" + highscore).slice (-8));
}

function tile (x, y) {
    return $("#f_"+y+"_"+x);
}

function show (t, what) {
    if (what == null) {
	t.html ("");
    } else if (symbols.includes (what)) {
	t.text (what);
    } else {
	t.text ("");
	t.append ($("<img/>", {src: what+".png"}));
    }
}

function redraw (x, y) {
    let t = tile (x, y);

    t.data ("actor", null);
    if (t.data ("piece")) {
	t.data ("piece").show ();
    } else {
	t.html ("");
    }
}

function move_x (x, dir) {
    return (size_x + x + dirs [dir][1]) % size_x;
}

function move_y (y, dir) {
    return (size_y + y + dirs [dir][0]) % size_y;
}

// TODO: update to use tile().data("walls") and tile.data("doors")? Caution: might be used before these exist!
// TODO: rename this.
function wall (level, theme, x, y) {
    let key = 0;
    // let my_tile = tile (x, y);

    if (x > 0 && wallchars.includes (level [y  ][x-1])) {
	key += 1;
    }
    if (y > 0 && wallchars.includes (level [y-1][x  ])) {
	key += 2;
    }
    if (x+1 < size_x && wallchars.includes (level [y  ][x+1])) {
	key += 4;
    }
    if (y+1 < size_y && wallchars.includes (level [y+1][x  ])) {
	key += 8;
    }

    key = key.toString (16);
    if (level [y][x] == "d") {
	key = "p"+key;
    } else if (theme) {
	key = theme[0]+key;
    }
    return key;
}

// Design decision: don't delve too deep into JavaScript's class concept. There be dragons.
class Piece {
    constructor (x, y, data) {
	this.x = x;
	this.y = y;
	this.type = data.type;
	this.pic = data.pic;
	this.value = data.value;
	this.power = data.power;

	tile (x, y).data ("piece", this);
	this.show ();
    }

    show () {
	show (tile (this.x, this.y), this.pic);
    }
};


class Monster {
    constructor (xPos, yPos, pic, ai, speed) {
	this.pic = pic;
	this.startX = xPos;
	this.startY = yPos;
        this.xPos = null;
        this.yPos = null;
        this.dir = null;
	this.speed = speed;
	this.ai = ai;
	this.ticks = 0;
	this.stuck = 0;
	this.is_dead = false;

	this.place (xPos, yPos);
    }

    remove () {
	redraw (this.xPos, this.yPos);
        this.xPos = null;
        this.yPos = null;
    }

    reset () {
	this.remove ();
	this.place (this.startX, this.startY);
    }

    show (t) {
	if (smileMan.empowered && (smileMan.empowered > 10 || (smileMan.empowered % 2))) {
	    show (t, this.is_dead ? "👀" : "👻");
	} else {
	    show (t, this.pic);
	}
    }

    choices () {
	let t = tile (this.xPos, this.yPos);
	if (t.data ("doors")) {
	    return [].concat (t.data ("passages"), t.data ("doors"));
	} else {
	    return t.data ("passages");
	}
    }

    place (x, y) {
	let t = tile (x, y);
	if (!this.is_dead && t.data ("actor")) {
	    this.stuck++;
	    return;
	} else {
	    this.stuck = 0;
	}

	if (this.xPos!=null && this.yPos!=null) {
	    let tt = tile (this.xPos, this.yPos);
	    if (this.is_dead && tt.data("actor")) {
		tt.data("actor").show (tt);
	    } else {
		redraw (this.xPos, this.yPos);
	    }
	}

        this.xPos = x;
        this.yPos = y;

	this.show (t);
	if (!this.is_dead) {
	    t.data ("actor", this);
	}
    }

    animate () {
	let speed = this.speed;
	if (smileMan.empowered && !this.is_dead) {
	    switch (difficulty) {
		case "easy": speed = Math.ceil (speed*1.75); break;
		case "normal": speed = Math.ceil (speed*1.3); break;
		case "hard": speed = Math.ceil (speed*1.1); break;
	    }
	}
	if (this.ticks < speed) {
	    this.show (tile (this.xPos, this.yPos));
	    this.ticks++;
	    return;
	}
	this.ticks = 0;

	if (this.is_dead) {
	    let pw = pathways [this.yPos][this.xPos][this.startY][this.startX];
	    this.dir = pw [0];
	} else if (smileMan.empowered) {
	    for (let dir of this.choices ()) {
		let x2 = move_x (this.xPos, dir);
		let y2 = move_y (this.yPos, dir);
		if (distance ({xPos: x2, yPos: y2}, smileMan) > distance (this, smileMan)) {
		    this.dir = dir;
		    break;
		}
	    }
	} else {
	    if (distance (this, smileMan) <= 1) {
		death ();
		return;
	    }
	    if (this.xPos == null || this.yPos == null) {
		this.stuck++;
	    }
	    if (this.stuck > 7) {
		this.reset ();
		return;
	    }
	    this.ai (this);
	}
	if (!this.dir) {
	    return;
	}

	let walls = tile (this.xPos, this.yPos).data ("walls");
	if (!walls || !walls.includes (this.dir)) {
	    this.place (move_x (this.xPos, this.dir), move_y (this.yPos, this.dir));
	}
    }
}


class FireManager {
    constructor (speed, pct, pic, scared, startPositions) {
	this.pic = pic;
	this.scared = scared;
	this.startPositions = startPositions;
	this.speed = speed;
	this.pct = pct;
	this.ticks = 0;
	the_flames = this;
    }

    show () {
	show ($(".flame"), this.pic);
    }

    ignite (t) {
	let p = t.data ("piece");
    	if (p && (p.type == "food" || p.type == "power")) {
	    toEat--;
	}
	t.addClass ("flame");
	t.data ("piece", null);
	t.data ("actor", this);
	if (toEat <= 0) {
	    this.show ();
	    level_done ();
	}
    }

    reset () {}

    extinguish (t) {
	t.removeClass ("flame");
	t.data ("actor", null);
	t.data ("piece", null);
	t.html ("");
    }

    spread () {
	$.each ($(".flame"), function (index, value) {
	    let choices = [];
	    let neighbors = $(value).data ("neighbors");
	    for (dir of $(value).data ("passages")) {
		let t = tile (neighbors [dir].x, neighbors [dir].y);
		if (t.data ("actor") == smileMan) {
		    death ();
		    return;
		} else if (!t.data ("actor")) {
		    choices.push (t);
		}
	    }
	    if (choices.length && pct (the_flames.pct)) {
		the_flames.ignite (choice (choices));
	    }
	});
    }

    animate () {
	let speed = this.speed;
	if (this.ticks < speed) {
	    this.show ();
	    this.ticks++;
	    return;
	}
	this.ticks = 0;

	if ($(".flame").length) {
	    this.spread ();
	    if (!pct(this.pct) || !pct(this.pct)) {
		return;
	    }
	}

	let pos = choice (this.startPositions);
	let t = tile (pos.x, pos.y);
	if (!t.data ("actor")) {
	    this.ignite (t);
	}
    }
}

function death () {
    if (gameTimer) {
	clearInterval (gameTimer);
	gameTimer = null;
    }
    if (smileMan.state != "default") {
	console.log ("Prevent overkill.");
	return;
    }
    current_tune.pause ();
    smileMan.update ("panic");
    $("#lives").text (smileys ["panic"].repeat (lives-1));
    effect ("caught", death2, 1000);
}

function death2 () {
    smileMan.update ("dead1");
    $.each (active_monsters, function (index, monster) {
	monster.reset ();
    });
    effect ("death", death3, 1000);
}

function death3 () {
    smileMan.update ("dead2");
    // effect ("death2", retry, 1000);
    setTimeout (retry, 1000);
}

function lose_all_lives () {
    lives--;
    if (lives > 0) {
	$("#lives").text (smileys ["cool"].repeat (lives-1));
	effect ("caught", lose_all_lives, 1000);
    } else {
	you_lost ();
    }
}

function you_lost () {
    $("#winorlose").text ("YOU HAVE LOST!");
    $("#drama").text ("💀");
    $(".gameover").show ();
}

function retry () {
    lives--;
    if (lives > 0) {
	$("#lives").text (smileys ["cool"].repeat (lives-1));
	smileMan.reset ();
	if (music) {
	    current_tune.volume = music_volume/100;
	    current_tune.play ();
	}
	if (!gameTimer) {
	    gameTimer = setInterval(game_loop, baseSpeed);
	}
	// ticks = 0;
    } else {
	you_lost ();
    }
}

function level_done () {
    if (smileMan.state != "default") {
	// Pure paranoia!
	return;
    }

    current_tune.pause ();
    $("#lives").text (smileys ["party"].repeat (lives-1));
    smileMan.update ("cool");
    if (gameTimer) {
	clearInterval (gameTimer);
	gameTimer = null;
    }
    $.each (active_monsters, function (index, monster) {
	monster.reset ();
    });
    active_monsters = [];

    effect ("level");

    rofl_count = 0;
    rofl_timer = setInterval (rofl, 400);
}


function rofl () {
    if (rofl_count % 2) {
	smileMan.update ("rofl1");
    } else {
	smileMan.update ("rofl2");
    }

    rofl_count++;
    if (rofl_count > 10) {
	// animate this?!?
	add_score (400*(1+current_level)*lives*lives);
	clearInterval (rofl_timer);
	current_level++;
	if (current_level < nrLevels) {
	    $("#extra").hide ();
	    $("#continue").hide ();
	    $("#lives2").text (smileys ["party"].repeat (lives));
	    $(".cleared").show ();
	    setTimeout (extra_live, 500);
	} else {
	    effect ("victory");
	    $("#winorlose").text ("You are the WINNER!");
	    $("#drama").text ("🖖");
	    $(".gameover").show ();
	}
    }
}

function extra_live () {
    effect ("extra");
    lives++;
    $("#lives2").text (smileys ["party"].repeat (lives));
    $("#extra").show ();
    $("#continue").show ();
}

class Smiley {
    constructor (paras) {
	this.startX = paras.x;
	this.startY = paras.y;
	this.empowered = 0;
	this.fast = 0;
	this.rampage = 0;
        this.xPos = null;
        this.yPos = null;
        this.startDir = paras.dir;
        this.dir = paras.dir;
        this.state = "default";
	this.speed = paras.speed;
	this.ticks = 0;

	this.place (paras.x, paras.y);
    }

    remove () {
	redraw (this.xPos, this.yPos);
        this.xPos = null;
        this.yPos = null;
	this.empowered = 0;
	this.fast = 0;
    }

    reset () {
	this.remove ();
	this.state = "default";
	this.dir = this.startDir;
	this.place (this.startX, this.startY);
    }

    show () {
	show (tile (this.xPos, this.yPos), smileys [this.state]);
    }

    place (x, y) {
	let t = tile (x, y);
	if (t.data ("actor")) {
	    if (t.data ("actor") == the_flames) {
		lose_all_lives ();
	    } else if (!this.empowered) {
		return;
	    } else {
		let mon = t.data ("actor");
		if (!mon.is_dead) {
		    effect ("eat_monster");
		    mon.is_dead = true;
		    this.rampage++;
		    add_score (100*(1+this.rampage*this.rampage));
		}
	    }
	}
	if (this.xPos!=null && this.yPos!=null) {
	    redraw (this.xPos, this.yPos);
	}

        this.xPos = x;
        this.yPos = y;
	let p = t.data ("piece");
	let pic = this.state;
    	if (p && (p.type == "food" || p.type == "power")) {
	    add_score (p.value);
	    t.data ("piece", null);
	    if (p.type == "power") {
		this.empowered = p.power;
		this.fast = p.speedup;
		effect ("power");
	    } else {
		if (p.value == 10) {
		    effect ("eat");
		    // restart_effect ("eat");
		} else {
		    effect ("fruit");
		}
	    }
	    pic = "fed";
	    toEat--;
	}
	show (t, smileys [pic]);
	t.data ("actor", this);
	if (toEat <= 0) {
	    level_done ();
	}
    }

    update (state=null) {
	if (state) {
	    this.state = state;
	}
	show (tile (this.xPos, this.yPos), smileys [this.state]);
    }

    animate () {
	if (!this.dir) {
	    return;
	}
	if (this.ticks < this.speed) {
	    this.ticks++;
	    if (this.fast) {
		this.ticks++;
	    }
	    return;
	}
	this.ticks = 0;
	if (this.empowered) {
	    this.empowered -= 1;
	    if (!this.empowered) {
	        this.rampage = 0;
		for (let monster of active_monsters) {
		    monster.is_dead = false;
		}
	    }
	}
	if (this.fast) {
	    this.fast -= 1;
	}

	if (tile (this.xPos, this.yPos).data ("passages").includes (this.dir)) {
	    this.place (move_x (this.xPos, this.dir), move_y (this.yPos, this.dir));
	}
    }
}


function game_loop () {
    ticks++;
    if (events [ticks]) {
	ev = events [ticks];
	switch (ev.type) {
	    case "monster":
		let ai = monster_types [ev.sub].ai;
		if (ev.ai) {
		    ai = ev.ai;
		}
		active_monsters.push (new Monster (ev.x, ev.y, monster_types [ev.sub].pic, ais [ai], ev.speed));
		effect ("monster");
		break;
	    case "fire":
		active_monsters.push (new FireManager (ev.speed, ev.pct, "🔥", "fire", ev.startPositions));
		// effect ("fire");
		effect ("monster");
		break;
	    case "item":
		effect ("bonus");
		let bonus = bonus_types [ev.sub];
		if (!bonus) {
		    bonus = choice (Object.values (bonus_types));
		}
		let p = tile (ev.x, ev.y).data ("piece");
		if (!p || (p.type != "food" && p.type != "power")) {
		    toEat++;
		}
		new Piece (ev.x, ev.y, bonus); // "b", bonus.type, bonus.value, bonus.pic);
		break;
	}
    }
    smileMan.animate ();
    for (let monster of active_monsters) {
	monster.animate ();
    };
}


function classify_neighbor (level, x, y, dir) {
    switch (level.map [move_y (y, dir)][move_x (x, dir)]) {
	case "w":
	    return "walls";
	case "d":
	    return "doors";
	default:
	    return "passages";
    }
}

function calculate_pathways (map) {
    let i, x1, y1, x2, y2;

    for (y1 = 0; y1 < size_y; y1++) {
	for (x1 = 0; x1 < size_x; x1++) {
	}
    }

    let last_run = [];

    for (y1 = 0; y1 < size_y; y1++) {
	pathways [y1] = Array(size_x);
	for (x1 = 0; x1 < size_x; x1++) {
	    pathways [y1][x1] = Array(size_y);
	    for (y2 = 0; y2 < size_y; y2++) {
		pathways [y1][x1][y2] = Array(size_x);
	    }

	    if (map [y1][x1] != "w") {
		reachable.push ([x1,y1]);
		pathways [y1][x1][y1][x1] = [null, 0];
		for (let dir in dirs) {
		    x2 = move_x (x1, dir);
		    y2 = move_y (y1, dir);
		    if (map [y2][x2] != "w") {
			pathways [y1][x1][y2][x2] = [dir, 1];
			last_run.push ([y1, x1, y2, x2, dir, 1]);
		    }
		}
	    }
	}
    }

    let j = 1;
    while (last_run.length) {
	j++;
	new_run = [];
	for (i of last_run) {
	    y1 = i [0];
	    x1 = i [1];
	    y2 = i [2];
	    x2 = i [3];
	    let old_dir = i [4]
	    let d = i [5];
	    for (dir in dirs) {
		let x3 = move_x (x2, dir);
		let y3 = move_y (y2, dir);
		if (map [y3][x3] != "w" && !pathways [y1][x1][y3][x3]) {
		    pathways [y1][x1][y3][x3] = [old_dir, d+1];
		    new_run.push ([y1, x1, y3, x3, old_dir, d+1]);
		}
	    }
	}
	last_run = new_run;
    }
}

function new_level (start_timer=true) {
    $(".tile").hide ();
    let level = levels [current_level];

    $("#level").text ("Level: "+(current_level+1));
    $(".flame").removeClass ("flame");
    the_flames = null;
    $(".tile").text ("");
    $(".tile").data ("piece", null);
    $(".tile").data ("walls", null);
    $(".tile").data ("passages", null);
    $(".tile").data ("doors", null);
    $(".tile").data ("actor", null);
    $("#lives").text (smileys ["cool"].repeat (lives-1));

    if (level.theme == "fire") {
	$("body").addClass ("fire");
    } else {
	$("body").removeClass ("fire");
    }

    toEat = 0;

    for (let y = 0; y < size_y; y++) {
	for (let x = 0; x < size_x; x++) {
            let piece = mappiece_types [level.map [y][x]];

	    if (piece) {
		if (piece.type == "wall") {
		    piece.pic = wall (level.map, level.theme, x, y);
		} else {
		    toEat++;
		}
		new Piece (x, y, piece);
	    }

	    let neighbors = {}
	    for (let dir in dirs) {
		neighbors [dir] = {x: move_x (x, dir), y: move_y (y, dir)};
		let neighbor_class = classify_neighbor (level, x, y, dir);
		let temp = tile (x, y).data (neighbor_class);
		if (temp) {
		    temp.push (dir);
		} else {
		    temp = [dir];
		}
		tile (x, y).data (neighbor_class, temp);
	    };
	    tile (x, y).data ("neighbors", neighbors);
	}
    }
    calculate_pathways (level.map);

    // toEat = 10;

    active_monsters = [];
    smileMan = new Smiley (level.startAt);
    ticks = 0;
    events = level.events;
    $(".tile").show ();

    current_tune = music_objects [level.tune];
    if (!current_tune) {
	current_tune = choice (Object.values (music_objects));
    }
    if (music) {
	current_tune.volume = music_volume/100;
	current_tune.currentTime = 0;
	current_tune.play ();
    }
    if (start_timer && !gameTimer) {
	gameTimer = setInterval(game_loop, baseSpeed);
    }
}


function key_handler (ev) {
    // ev.preventDefault();
    switch (ev.keyCode) {
	case 37: smileMan.dir = "w"; break;
	case 38: smileMan.dir = "n"; break;
	case 39: smileMan.dir = "e"; break;
	case 40: smileMan.dir = "s"; break;
    }
}

function toggle_music (value=null) {
    if (value=="false" || value==false) {
	music = false;
    } else if (value=="true" || value==true) {
	music = true;
    } else {
	music = !music;
    }
    if (music) {
	current_tune.volume = music_volume/100;
	current_tune.play ();
	$("#music").text ("🎶🔇");
    } else {
	current_tune.pause ();
	$("#music").text ("🎶🔊");
    }
    window.localStorage.setItem ("smiling_music_on", music);
}

function toggle_sound (value=null) {
    if (value=="false" || value==false) {
	sound = false;
    } else if (value=="true" || value==true) {
	sound = true;
    } else {
	sound = !sound;
    }
    if (sound) {
	$("#sound").text ("🔔🔇");
    } else {
	$("#sound").text ("🔔🔊");
    }
    window.localStorage.setItem ("smiling_sound_on", sound);
}

function init_sound () {
    $.each (tunes, function (index, value) {
	let the_tune = $("<audio>", {src:value})[0];
	the_tune.addEventListener("ended", function() {this.volume = music_volume/100; this.play();}, false);
	music_objects[index] = the_tune;
    });
    $.each (sounds, function (index, value) {
	sound_objects[index] = $("<audio>", {src:value})[0];
    });
}

function set_music_volume (val) {
    music_volume = val;
    if (current_tune) {
	current_tune.volume = val/100;
    }
    window.localStorage.setItem ("smiling_music_volume", music_volume);
}

function set_sound_volume (val) {
    sound_volume = val;
    window.localStorage.setItem ("smiling_sound_volume", sound_volume);
}

$(function () {
    let x, y, row;

    init_sound ();

    $(this).keydown(key_handler);

    init_score ();

    for (y = 0; y < size_y; y++) {
	for (x = 0; x < size_x; x++) {
	    $("#board").append ($("<div/>", {class: "tile", id:"f_"+y+"_"+x}));
	}
    }

    $(".help").click (function () {window.location.href = "help.html";});
    $(".about").click (function () {window.location.href = "about.html";});

    difficulty = window.localStorage.getItem ("smiling_difficulty");
    if (!["easy", "normal", "hard"].includes (difficulty)) {
	difficulty = "normal";
    }
    window.localStorage.setItem ("smiling_difficulty", difficulty);
    $("input[name=difficulty][value='"+difficulty+"']").prop("checked",true);
    $("input[name=difficulty]").click (function (ev) {
	difficulty = ev.currentTarget.value;
    });

    $("#left").click (function () {smileMan.dir = "w";});
    $("#right").click (function () {smileMan.dir = "e";});
    $("#up").click (function () {smileMan.dir = "n";});
    $("#down").click (function () {smileMan.dir = "s";});

    music_volume = window.localStorage.getItem ("smiling_music_volume");
    sound_volume = window.localStorage.getItem ("smiling_sound_volume");
    if (music_volume == null) {
	set_music_volume (50);
    }
    if (sound_volume == null) {
	set_sound_volume (100);
    }
    $("#music_volume").val (music_volume);
    $("#sound_volume").val (sound_volume);
    $("#music_volume").change (function (ev) {set_music_volume (ev.target.value);});
    $("#sound_volume").change (function (ev) {set_sound_volume (ev.target.value);});
    $("#music").click (toggle_music);
    $("#sound").click (toggle_sound);
    $("#continue").click (function (ev) {$(".cleared").hide (); new_level();});

    new_level (false);

    $("#start").click (function () {
	baseSpeed = {
	    easy: 45,
	    normal: 30,
	    hard: 17} [difficulty];
	score_factor = {
	    easy: 0.5,
	    normal: 1.0,
	    hard: 1.5} [difficulty];
	window.localStorage.setItem ("smiling_difficulty", difficulty);
	$(".modal").hide ();
	toggle_sound (window.localStorage.getItem ("smiling_sound_on"));
	toggle_music (window.localStorage.getItem ("smiling_music_on"));
	if (!gameTimer) {
	    gameTimer = setInterval(game_loop, baseSpeed);
	}
    });
    $("#restart").click (function () {
	window.location.href = "index.html";
    });
});
